/*	
 *	clipmp3.c
 *
 *	P/ECE MP3 Driver
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2003 Naoyuki Sawa
 *
 *	* Tue Oct 28 08:11:00 JST 2003 Naoyuki Sawa
 *	- 1st [XB
 *	* Tue Nov  4 00:00:00 JST 2003 Naoyuki Sawa
 *	- SCMPXŃGR[hMP3t@CĐł悤AłB
 *	  ̐ݒŃGR[ht@CĐāAnOAbv͂ȂȂ܂B
 *		C[	III
 *		rbg[g	64000 Œ
 *		m
 *		Gt@VX	Ȃ
 *	  ͂܂܂ŁAƂĂmCY܂B
 *	  ́A܂V[gubNE~bNXubNɑΉĂȂłB
 *	  _ł́AǂOubNƂ݂ȂčĐĂ܂Ă܂B
 *	* Thu Nov  6 23:50:00 JST 2003 Naoyuki Sawa
 *	- V[gubNΉB
 *	- XP[t@N^ΉB
 *	- mp3_read_granule_info()̃oOCB
 *	- mp3_read_scalefactor()̃oOCB
 */
#include "clip.h"
#include "clipmp3.h" /* TODO:Cuɑgݍ񂾂O܂ */

/****************************************************************************
 *	t[wb_
 ****************************************************************************/

/* t[wb_ǂݍ݂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	frame		t[wb_i[\́B
 * [out]
 *	߂l		Ȃ0Ԃ܂B
 *			sȂ0ȊO̒lԂ܂B
 */
int
mp3_read_frame_header(MP3BITSTREAM* bs, MP3FRAMEHEADER* frame)
{
	frame->syncword			= mp3_bitstream_get(bs, 12);
	frame->id			= mp3_bitstream_get(bs,  1);
	frame->layer			= mp3_bitstream_get(bs,  2);
	frame->protection_bit		= mp3_bitstream_get(bs,  1);
	frame->bitrate_index		= mp3_bitstream_get(bs,  4);
	frame->sampling_frequency	= mp3_bitstream_get(bs,  2);
	frame->padding_bit		= mp3_bitstream_get(bs,  1);
	frame->private_bit		= mp3_bitstream_get(bs,  1);
	frame->mode			= mp3_bitstream_get(bs,  2);
	frame->mode_extention		= mp3_bitstream_get(bs,  2);
	frame->copyright		= mp3_bitstream_get(bs,  1);
	frame->original			= mp3_bitstream_get(bs,  1);
	frame->emphasis			= mp3_bitstream_get(bs,  2);

	/* Œ̃G[`FbNB */
	if(frame->syncword != 0xfff) return -1;	/* [h */
	if(frame->id       !=     1) return -1;	/* MPEG1      */
	if(frame->layer    !=     1) return -1;	/* LayerIII   */

	/* CRCی삠ȂACRCیrbgǂݔ΂܂B
	 * uprotection_bit=0:CRCی삠/1:ȂvłBv!!
	 */
	if(!frame->protection_bit) mp3_bitstream_skip(bs, 16);

	return 0;
}

/****************************************************************************
 *	Oj[
 ****************************************************************************/

/* Oj[ǂݍ݂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	granule		Oj[i[\́B
 */
void
mp3_read_granule_info(MP3BITSTREAM* bs, MP3GRANULEINFO* granule)
{
	granule->part2_3_length			= mp3_bitstream_get(bs, 12);
	granule->big_values			= mp3_bitstream_get(bs,  9);
	granule->global_gain			= mp3_bitstream_get(bs,  8);
	granule->scalefac_compress		= mp3_bitstream_get(bs,  4);
	granule->window_switching_flag		= mp3_bitstream_get(bs,  1);
	if(!granule->window_switching_flag) {
		granule->block_type		= 0;
		granule->table_select[0]	= mp3_bitstream_get(bs,  5);
		granule->table_select[1]	= mp3_bitstream_get(bs,  5);
		granule->table_select[2]	= mp3_bitstream_get(bs,  5);
		granule->region0_count		= mp3_bitstream_get(bs,  4);
		granule->region1_count		= mp3_bitstream_get(bs,  3);
	} else {
		granule->block_type		= mp3_bitstream_get(bs,  2);
		granule->mixed_block_flag	= mp3_bitstream_get(bs,  1);
		granule->table_select[0]	= mp3_bitstream_get(bs,  5);
		granule->table_select[1]	= mp3_bitstream_get(bs,  5);
		granule->subblock_gain[0]	= mp3_bitstream_get(bs,  3);
		granule->subblock_gain[1]	= mp3_bitstream_get(bs,  3);
		granule->subblock_gain[2]	= mp3_bitstream_get(bs,  3);
		if(granule->block_type == 2 && !granule->mixed_block_flag) {
			granule->region0_count	=       8         ;
			granule->region1_count	= 22 - (8 + 1) - 1;
		} else {
			granule->region0_count	=       7         ;
			granule->region1_count	= 22 - (7 + 1) - 1;
		}
	}
	granule->preflag			= mp3_bitstream_get(bs,  1);
	granule->scalefac_scale			= mp3_bitstream_get(bs,  1);
	granule->count1table_select		= mp3_bitstream_get(bs,  1);
}

/****************************************************************************
 *	TCh
 ****************************************************************************/

/* TChǂݍ݂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	side		TChi[\́B
 *	channels	`lB
 */
void
mp3_read_side_info(MP3BITSTREAM* bs, MP3SIDEINFO* side, int channels)
{
	int i_granule;
	int i_channel;

	side->main_data_begin		= mp3_bitstream_get(bs, 9);
	switch(channels) {
	case 1:
		side->private_bits	= mp3_bitstream_get(bs, 5);
		side->scfsi[0]		= mp3_bitstream_get(bs, 4);
		break;
	case 2:
		side->private_bits	= mp3_bitstream_get(bs, 3);
		side->scfsi[0]		= mp3_bitstream_get(bs, 4);
		side->scfsi[1]		= mp3_bitstream_get(bs, 4);
		break;
	}
	for(i_granule = 0; i_granule < 2; i_granule++) {
		for(i_channel = 0; i_channel < channels; i_channel++) {
			mp3_read_granule_info(bs, &side->granule_info[i_granule][i_channel]);
		}
	}
}

/****************************************************************************
 *	XP[t@N^
 ****************************************************************************/

/* XP[t@N^ǂݍ݂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	granule		Oj[\́B
 *	scale		XP[t@N^[\́B(ɃXP[t@N^[ǂݍ݂܂)
 *	scfsi		1Oj[̏ꍇA0nĂB
 *			2Oj[̏ꍇAMP3SIDEINFO.scsfi[ch#]̒lnĂB
 * [out]
 *	*scale		XP[t@N^[ǂݍ݁Ai[܂B
 * [note]
 *	* 2Oj[̓ǂݍ݂ł́Aꕔ̒l1Oj[ƋLꍇ܂B
 *	  ̏ꍇAω镔ǂݍ݁AL͓ǂݍ݂܂B
 *	  ]āA1Oj[œǂݍ񂾃XP[t@N^[\̂A
 *	  ej󂹂ɓ`l̑2Oj[̃XP[t@N^[ǂݍ݂ɎgĂB
 */
void
mp3_read_scalefactor(MP3BITSTREAM* bs, const MP3GRANULEINFO* granule, MP3SCALEFACTOR* scale, int scfsi)
{
	int i, slen1, slen2;
	short* out;

	slen1 = mp3_scalefactor_compress_table[granule->scalefac_compress][0];
	slen2 = mp3_scalefactor_compress_table[granule->scalefac_compress][1];

	if(granule->block_type == 2) {
		if(granule->mixed_block_flag) {
			/* ~bNXubN */
			out = &scale->long_block[(i = 0)];
			do { *out++ = mp3_bitstream_get(bs, slen1); } while(++i <= 7);	/* long_block[0`7] */
			out = &scale->short_block[(i = 3)][0];
			do { *out++ = mp3_bitstream_get(bs, slen1);
			     *out++ = mp3_bitstream_get(bs, slen1);
			     *out++ = mp3_bitstream_get(bs, slen1); } while(++i <= 5);	/* short_block[3` 5] */
		} else {
			/* V[gubN */
			out = &scale->short_block[(i = 0)][0];
			do { *out++ = mp3_bitstream_get(bs, slen1);
			     *out++ = mp3_bitstream_get(bs, slen1);
			     *out++ = mp3_bitstream_get(bs, slen1); } while(++i <= 5);	/* short_block[0` 5] */
		}
		do { *out++ = mp3_bitstream_get(bs, slen2);
		     *out++ = mp3_bitstream_get(bs, slen2);
		     *out++ = mp3_bitstream_get(bs, slen2); } while(++i <= 11);		/* short_block[6`11] */
		     *out++ = 0;
		     *out++ = 0;
		     *out++ = 0;							/* short_block[   12] */
	} else {
		/* OubN */
		if(!(scfsi & 1 << 3)) { for(i =  0; i <=  5; i++) { scale->long_block[ i] = mp3_bitstream_get(bs, slen1); } };	/* long_block[ 0` 5] */
		if(!(scfsi & 1 << 2)) { for(i =  6; i <= 10; i++) { scale->long_block[ i] = mp3_bitstream_get(bs, slen1); } };	/* long_block[ 6`10] */
		if(!(scfsi & 1 << 1)) { for(i = 11; i <= 15; i++) { scale->long_block[ i] = mp3_bitstream_get(bs, slen2); } };	/* long_block[11`15] */
		if(!(scfsi & 1 << 0)) { for(i = 16; i <= 20; i++) { scale->long_block[ i] = mp3_bitstream_get(bs, slen2); } };	/* long_block[16`20] */
								    scale->long_block[21] = 0;					/* long_block[    21] */
	}
}

/****************************************************************************
 *	MP3hCo
 ****************************************************************************/

#define MP3BUFLEN	(64 * 5)

int
mp3_init(MP3DRIVER* mp3, const void* data, int len)
{
	/* ܂[NAB */
	memset(mp3, 0, sizeof(MP3DRIVER));

	/* ǂݍ݈ʒuƏI[ʒui[܂B */
	mp3->pos = data;
	mp3->end = mp3->pos + len;

	/* ŏ̃t[fR[hAobt@𖞂܂B */
	if(mp3_decode_frame(mp3) != 0) return -1;

	return 0;
}

int
mp3_stream_callback(short* wbuff, int param)
{
	MP3DRIVER* mp3 = (MP3DRIVER*)param;
	short* wbuff_end = wbuff + MP3BUFLEN;
	int sampling_frequency = mp3->sampling_frequency;
	int i_fbuff = mp3->i_fbuff;
	int d_fbuff = mp3->d_fbuff;
	short* fbuff = mp3->fbuff;

	while(wbuff < wbuff_end) {
		*wbuff++ = fbuff[i_fbuff];
		d_fbuff += sampling_frequency;
		while(d_fbuff >= 16000) {
			d_fbuff -= 16000;
			i_fbuff++;
		}
		while(i_fbuff >= MP3_FRAME_SIZE_HALF) {
			if(mp3_decode_frame(mp3) != 0) return -1;
			i_fbuff -= MP3_FRAME_SIZE_HALF;
			sampling_frequency = mp3->sampling_frequency;
		}
	}

	mp3->i_fbuff = i_fbuff;
	mp3->d_fbuff = d_fbuff;

	return 0;
}

/* 1t[fR[h܂B
 * [in]
 *	mp3			MP3hCo\́B
 * [out]
 *	߂l			Ȃ0Ԃ܂B
 *				sȂ0ȊO̒lԂ܂B(I[ɒBꍇȂ)
 *	mp3->fbuff		fR[ht[f[^i[܂B
 *	mp3->sampling_frequency	fR[ht[̃TvOgi[܂B
 */
int
mp3_decode_frame(MP3DRIVER* mp3)
{
	MP3BITSTREAM bs, bs_tmp;
	MP3FRAMEHEADER frame;
	MP3SIDEINFO side;
	MP3SCALEFACTOR scale;
	MP3GRANULEINFO* granule;
	int bitrate, sampling_frequency, channels, frame_size, main_data_size;
	int i_granule, i_channel;
	short* out;

	/* t[ԂŕۑKv̂Ȃobt@B */
#ifndef PIECE
	/* SRAMɎꍇ */
	static short buf1[MP3_GRANULE_SIZE];
	static short buf2[MP3_GRANULE_SIZE];
#else
	/* RAMɎꍇ */
	static short* buf1 = (short*)0x2000 - MP3_GRANULE_SIZE * 1;
	static short* buf2 = (short*)0x2000 - MP3_GRANULE_SIZE * 2;
	extern unsigned long __END_FRAM_ADDR[];
	if((int)buf2 < (int)__END_FRAM_ADDR) DIE();
#endif

	/* t[Jnʒu܂B */
	for(;;) {
		/* f[^I[? */
		if(mp3->pos >= mp3->end - 1) return -1;
		/* [h? */
		if(mp3->pos[0] == 0xff &&
		  (mp3->pos[1] &  0xf0) == 0xf0) break;
		mp3->pos++;
	}

	/* rbgXg[J܂B */
	mp3_bitstream_open(&bs, mp3->pos);

	/* t[wb_ǂݍ݂܂B */
	if(mp3_read_frame_header(&bs, &frame) != 0) return -1/*t[wb_s*/;
	bitrate = mp3_bitrate_table[frame.bitrate_index];
	sampling_frequency = mp3_sampling_frequency_table[frame.sampling_frequency];
	channels = frame.mode == 3 ? 1 : 2;
	frame_size = 144 * bitrate / sampling_frequency + frame.padding_bit; /* t[TCYvZ */

	/* TChǂݍ݂܂B */
	mp3_read_side_info(&bs, &side, channels);

	/* Cf[^̃rbg~ρB */
	ASSERT(bs.cnt == 0); /* TCh̖ŃoCgEɐ񂵂Ă͂ */
	main_data_size = frame_size - (bs.pos - mp3->pos)/*t[wb_ƃTChc肪Cf[^*/;
	if(mp3->main_save_size + main_data_size > sizeof mp3->main_save) {
		memmove(&mp3->main_save[0], &mp3->main_save[mp3->main_save_size - side.main_data_begin], side.main_data_begin);
		mp3->main_save_size = side.main_data_begin;
	}
	memcpy(&mp3->main_save[mp3->main_save_size], bs.pos, main_data_size);
	mp3_bitstream_open(&bs, &mp3->main_save[mp3->main_save_size - side.main_data_begin]);
	mp3->main_save_size += main_data_size;

	/* eOj[fR[h܂B */
	out = mp3->fbuff;
	for(i_granule = 0; i_granule < MP3_GRANULE_COUNT; i_granule++) {
		for(i_channel = 0; i_channel < channels; i_channel++) {
			granule = &side.granule_info[i_granule][i_channel];
			if(i_channel == 0) { /* `l̂ݎgp */
				bs_tmp = bs;
				/* XP[t@N^ǂݍ݁B */
				mp3_read_scalefactor(&bs_tmp, granule, &scale, !i_granule ? 0 : side.scfsi[i_channel]);
				/* nt}B */
				mp3_huffman_decode(&bs_tmp, &frame, granule, buf1);
				if(granule->block_type == 2) {
					if(granule->mixed_block_flag) {
						/* ~bNXubN */
						/* TODO: ~bNXubN𐶐GR[_܂݂Ȃ炵̂ŁAƂ肠KvȂ悤łB */
						memset(buf1, 0, sizeof buf1);
					} else {
						/* V[gubN */
						/* tʎqB */
						mp3_requantize_short(&frame, granule, &scale, buf1, buf2);
						/* V[gubNɂ͂ƂƃGAVO팸͂܂B */
						/* IMDCTB */
						mp3_imdct_short(buf2, buf1, &mp3->imdct_save[0]);
					}
				} else {
					/* OubN */
					/* tʎqB */
					mp3_requantize_long(&frame, granule, &scale, buf1, buf2);
					/* TODO: GAVO팸͖ΉłB */
					/* IMDCTB */
					mp3_imdct_long(buf2, buf1, &mp3->imdct_save[0]);
				}
				/* TuohB */
				mp3->subband_synthesys_offset = mp3_subband_synthesys(buf1, out, &mp3->subband_synthesys_save[0][0], mp3->subband_synthesys_offset);
				out += MP3_GRANULE_SIZE_HALF;
			}
			mp3_bitstream_skip(&bs, granule->part2_3_length);
		}
	}

	/* TvOgi[܂B */
	mp3->sampling_frequency = sampling_frequency / 2/*=HALF*/;

	/* t[ʒui߂܂B */
	mp3->pos += frame_size;

	return 0;
}

/****************************************************************************
 *	AvP[Vp֐
 ****************************************************************************/

int
mp3_play(const void* data, int len)
{
	static MP3DRIVER mp3; /* STATICł! */

	/* ܂mɒ~܂B */
	mp3_stop();

	/* MP3hCo܂B */
	if(mp3_init(&mp3, data, len) != 0) return -1;

	/* Xg[ĐJn܂B */
	stream_play(MP3BUFLEN, mp3_stream_callback, (int)&mp3);

	return 0;
}

void
mp3_stop()
{
	/* Xg[Đ~܂B */
	stream_stop();

	/* MP3hCõN[Abv͕svłB */
}

